#!/usr/bin/perl 
##########################################################################
#                                                                        #
# Perl Script for Radius Accounting                                      #
# Switzernet(c)2012                                                      #
#                                                                        #
##########################################################################
# my modules
use strict;
use warnings;
use Asterisk::Manager;
use Sys::Syslog;
use POSIX;
use Config::IniFiles;
use Authen::Radius;
use DBI;
use Switch;
use threads;
use threads::shared;

#Enable debug
use constant DEBUG => 0;
use Data::Dumper;
local $Data::Dumper::Indent = 0;

#WARNING Variables set by AGI are not correctly read from this script
##########################################################################
# Setting
my $lock_file = '/var/run/ast-rad-acc.pid';
my $lock_control = '/var/run/ast-control.pid';
my $runas_user = 'root';
my $monitor_dir = '/var/spool/asterisk/monitor';
my $config_dir = '/etc/astrad/config';

##########################################################################
# Globals
my %channels;
my ($ast_hostname,$ast_username,$ast_password);
my ($rad_serv, $rad_serv2, $rad_sec, $nas_ip);
my ($DB_NAME_Rad,$DB_HOST_Rad,$DB_USER_Rad,$DB_PASS_Rad,$dbh_Rad,$Rad_connected);
my ($CustomerType,$astman,$ast_connected,$thr);
my $t_event : shared;
my @event_list;

##########################################################################
#                             Initialization                             #
##########################################################################
# Check if already running
#
if( -e $lock_file ) {
  open(PID,$lock_file);
  my $pid=<PID>;
  close PID;
  chomp $pid;
  if( !-e "/proc/$pid" ) {
    print STDERR "Lock file present, but no process with pid=$pid.\n";
    die "Can't delete lock file $lock_file\n" if !unlink $lock_file;
    print STDERR "Lock file has been removed.\n";
  } else {
    die "Lockfile present, another copy is punning pid=$pid\n";
  }
}

load_config();
my ($name, $passwd, $uid, $gid) = getpwnam($runas_user) or die "$runas_user not in passwd file";;

##########################################################################
# set C locale
POSIX::setlocale(LC_ALL, "C");

##########################################################################
# Become daemon
my $pid;
if( !defined($pid = fork()) ) {
  die "cannot fork: $!";
} elsif ($pid) {
# Create lockfile, and finish parent process
#
  open(PID, "> $lock_file") || die "ast-rad-acc.pl: Unable to create lockfile $lock_file\n";
  print PID "$pid";
  close PID;
  chown $uid, $gid, $lock_file;
  exit;
} else {
# daemon
  setpgrp();
  select(STDERR); $| = 1;
  select(STDOUT); $| = 1;
  openlog('ast-rad-acc', 'cons,pid', 'daemon');
  syslog('notice', "Ast-Rad ast-rad-acc: accounting for Asterisk started");
}

##########################################################################
# Install signall handler
$SIG{INT} = \&safe_exit;
$SIG{QUIT} = \&safe_exit;
$SIG{TERM} = \&safe_exit;
$SIG{HUP} = \&load_config;

##########################################################################
# Drop privileges
setuid($uid);
$< = $uid;
$> = $uid;

##########################################################################
# Signal Handlers
sub safe_exit {
  my($sig) = @_;
  syslog('crit', "Ast-Rad: Caught a SIG$sig - shutting down");
  $astman->disconnect if $ast_connected;  
  $dbh_Rad->disconnect if $Rad_connected;    
  unlink $lock_file or syslog('crit', "Ast-Rad: Unable to create lockfile $lock_file\n");
  closelog();     
  exit;
}

##########################################################################
#                                FUNCTION                                #
##########################################################################
# Load config
sub load_config {
  my $conf=Config::IniFiles->new(-file => "$config_dir/switzer.conf");
  return 0 if !defined $conf ;
  $rad_serv     = $conf->val('MASTER_RD','RADIUS_SERVER');
  $rad_serv2    = $rad_serv;
  $rad_serv2    = $conf->val('MASTER_RD','RADIUS_SERVER2') if defined $conf->val('MASTER_RD','RADIUS_SERVER2');
  $rad_sec      = $conf->val('MASTER_RD','RADIUS_SECRET');
  $nas_ip       = $conf->val('GLOBAL','NAS_IP');    
  $DB_NAME_Rad  = $conf->val('ASTER_DB','DB_NAME');
  $DB_HOST_Rad  = $conf->val('ASTER_DB','DB_HOST');
  $DB_USER_Rad  = $conf->val('ASTER_DB','DB_USER');
  $DB_PASS_Rad  = $conf->val('ASTER_DB','DB_PASS');      
  $ast_hostname = $conf->val('Manager','AMI_PERMIT');
  $ast_username = $conf->val('Manager','AMI_USER2');
  $ast_password = $conf->val('Manager','AMI_SECRET2');
  return 1;
}
##########################################################################
# sample '09:16:05 GMT Sat Dec 11 2004'
sub format_date_gmt {
  my ($date) = @_;
  return strftime "%H:%M:%S GMT %a %b %e %Y", gmtime($date);
}

##########################################################################
# sample '2004-12-11 09:16:05'
sub format_date_sql {
  my ($date) = @_;
  if ($date eq 0) {
    return "NULL";
  } else {
    return strftime("%Y-%m-%d %H:%M:%S", localtime($date));
  }
}

##########################################################################
# Astrad Rad. DB connect 
sub Rad_connect {
  my $cop=0;
  $Rad_connected = 1;
  while( 1 ) {
    $cop++;
    $dbh_Rad = DBI->connect_cached("DBI:mysql:$DB_NAME_Rad;host=$DB_HOST_Rad", $DB_USER_Rad, $DB_PASS_Rad);
    if( $dbh_Rad ) {
      $Rad_connected = 1;
      $cop=0;
      syslog('info', 'Ast-Rad: Connected to Astrad Rad. DB!') if (DEBUG);
      return 1;
    } else {
      if ($Rad_connected) {
        syslog('err', 'Ast-Rad: Could not connect to Astrad Rad. DB!');
        $cop=1;
      }
      $Rad_connected = 0;
      syslog('info', "Ast-Rad: Connection to Astrad Rad. DB -> SLEEP $cop");
      $cop++ if $cop;
      return 0 if ($cop>=3);
    }
    sleep 1;
  }
}

##########################################################################
# Insert into the Failed radius packets Table 
sub Failed_radius_packet {
  my (%data) = @_;
  return 0 if (!Rad_connect());
  $data{'CDR(CustomerType)'}="" if (!defined($data{'CDR(CustomerType)'}));    
  my $request="INSERT INTO radius_packets_Failed ("
             ."h323_connect_time, h323_disconnect_time, acct_session_time, h323_disconnect_cause, nas_ip_address, user_name, calling_station_id, "
             ."called_station_id, call_id, h323_setup_time, h323_conf_id, h323_remote_address, call_profile, creation_date, reception_date, "
             ."last_send, send_counter, status, radius_ip) "
             ."VALUES "
             ."('".$data{'H323_connect_time'}."','".$data{'H323_disconnect_time'}."','".$data{'Acct_Session_Time'}."','".$data{'H323_disconnect_cause'}
             ."','".$data{'nas_ip_address'}."','".$data{'user_name'}."','".$data{'Calling-Station-Id'}."','".$data{'Called-Station-Id'}
             ."','".$data{'call_id'}."','".$data{'H323_setup_time'}."','".$data{'H323_conf_id'}."','".$data{'H323_remote_address'}."','".$data{'call_profile'}
             ."','".$data{'creation_date'}."','".$data{'reception_date'}."','".$data{'last_send'}."','".$data{'send_counter'}."','".$data{'status'}."','".$data{'radius_ip'}."')";
  my $sth = $dbh_Rad->prepare($request);
  if ($sth->execute()) {
    $sth->finish;
    syslog('info', "Ast-Rad: Request = $request") if (DEBUG);
    return 1;
  }
  return 0;
}
##########################################################################
# thread control
sub AMI_control {
  syslog('info', 'Ast-Rad: Thread: Start AMI Control') if (DEBUG);
  my ($l_event,$cop) = (0,0);
  while ( 1 ) {
    $l_event = time();
    syslog('info', "Ast-Rad: Thread: Time Events -> Last = $t_event | Now = $l_event") if (DEBUG);
    #TODO if connection to asterisk is lost the script should not exit
    if ($l_event - $t_event > 400) {
      syslog('crit', "Ast-Rad: Thread: Asterisk PBX is probably restarted !");
      safe_exit();
    }
    $cop=0 if ($l_event - $t_event < 4);
    $cop++;
    syslog('info', "Ast-Rad: Thread: AMI Control sleep $cop") if (DEBUG);
    sleep 2;
  }
}
##########################################################################
# AMI connect
sub AMI_connect {
  $astman = new Asterisk::Manager;
  $astman->user($ast_username);
  $astman->secret($ast_password);
  $astman->host($ast_hostname);
  $t_event = time();
  my $cop=0;
  $ast_connected = 1;
  while( 1 ) {
    $cop++;
    if( $astman->connect ) {
      $ast_connected = 1;
      $cop=0;
      syslog('info', 'Ast-Rad: Connected to Asterisk!') if (DEBUG);
      $thr = threads->new(\&AMI_control);
      $astman->setcallback('DEFAULT', \&status_callback);
      eval { $astman->eventloop; };
    } else { 
      if ($ast_connected) {
        syslog('err', 'Ast-Rad: Could not connect to Asterisk!');
        $cop=1;
      }
      $ast_connected = 0;
      syslog('info', "Ast-Rad: SLEEP $cop");
      $cop++ if $cop;
      return 0 if ($cop>=3);
    }
    sleep 1;
  }
}

##########################################################################
# 
sub status_callback {
  my (%unfiltered_event) = @_;
  my $status = 0;
  return if not (%unfiltered_event);
  $t_event = time();
  foreach (keys %unfiltered_event) {
    next if ($_ eq '');
    push(@event_list, [$_, $unfiltered_event{$_}]);
  }
  my @event_pos;
  for my $i ( 0 .. $#event_list) {
    push(@event_pos, $i) if ($event_list[$i][0] eq 'Event');
  }
  #Remove lines prior to first event
  my $offset=0;
  if ($event_pos[0] != 0) {
    shift(@event_list);
    $offset++;
  }
  while ($#event_pos > 0) {
    my @filtered_event;
    my %filtered_event_hash = ();
    my $eventname = '';
    my $addtolist = 1;
    CYCLE1: for (my $i = 0; $i < ($event_pos[1]-$offset) ; $i++) {
      print "['".$event_list[0][0]."','".$event_list[0][1]."']\n" if (DEBUG > 5);
      if ($addtolist) {
        if ($eventname eq '') {
          if ($event_list[0][1] eq 'PeerStatus') {
            $addtolist = 0;
          } else {
            $eventname = $event_list[0][0];
          }
        }
        push(@filtered_event, [$event_list[0][0],$event_list[0][1]]) if ($addtolist == 1);
        $filtered_event_hash{$event_list[0][0]}=$event_list[0][1];
      }
      shift(@event_list);
    }
    if (DEBUG > 4 && $eventname ne '') {
      while ( my ($key, $value) = each(%filtered_event_hash) ) {
        print "$key => $value\n";
      }
    }
    if ($eventname eq 'Event') {
      print "Executing event in filtered_event ".$filtered_event_hash{$eventname}."\n" if (DEBUG > 3);
      switch ($filtered_event_hash{$eventname}) {
        case "Newchannel" {
#Event => Newchannel
#Privilege => call,all
#Uniqueid => 91.121.122.64-1343980582.103
#Channel => SIP/41225504011-00000067
#Context => fromaccount
#CallerIDName => 41225504011
#ChannelState =>
#ChannelStateDesc => Down
#CallerIDNum => 41225504011

#Event => Newchannel
#Privilege => call,all
#Uniqueid => 91.121.122.64-1343980583.104
#Channel => SIP/212.249.15.5-00000068
#Context => fromhost
#ChannelState =>
#ChannelStateDesc => Down
#Value => 3ef7ad58361b3a74394a32b041f89ce7@91.121.122.64:5060
          my $ncuniqueid = $filtered_event_hash{Uniqueid};
          my $nccin      = '';
          $nccin         = $filtered_event_hash{CallerIDNum} if defined($filtered_event_hash{CallerIDNum});
          my $ncchannel  = $filtered_event_hash{Channel};
          my $context    = '';
          $context       = $filtered_event_hash{Context} if defined($filtered_event_hash{Context});
          if ($ncchannel ne '') {
            $channels{$ncchannel} = {
              'CALLERID' => $nccin,
              'UNIQUEID' => $ncuniqueid,
              'CALL_START' => time(),
              'LINK_START' => time(),
              'LINK_END' => time(),
              'STARTSENT' => 0,
              'CALL_TYPE' => 'VoIP',
              'CALL_PROTOCOL' => 'sipv2',
              'CALL_ID' => $ncuniqueid,
              'RADIUS_Server' => $rad_serv,
              'RADIUS_Secret' => $rad_sec,
              'NAS_IP_Address' => $nas_ip,
              'NAS-Port' => '5060',
              'XFOLLOW_ME' => '0',
              'CHANNEL1' => $ncchannel,
              'COUNT' => 1,
              'Remoteip' => $nas_ip,
              'DIALED' => 0,
              'ANSWERED' => 0,
              'CONTEXT' => $context
            };
            $channels{$ncchannel}{'Remoteip'} = $1 if ($ncchannel =~ /^SIP\/(\d+\.\d+\.\d+\.\d+)-/);
            if (DEBUG) {
              print 'Newchannel Set $channels{'.$ncchannel.'}{UNIQUEID} = '.$ncuniqueid."\n" if (DEBUG);
              print 'Total '.keys( %channels )." active channels.\n" if (DEBUG > 1);
            }
          }
        }
        case "Newstate" {
#['Event','Newstate']
#['Privilege','call,all']
#['Channel','SIP/41215500331-00000013']
#['CallerIDName','41215500331']
#['Uniqueid','91.121.122.64-1324372616.19']
#['ChannelStateDesc','Up']
#['ChannelState','6']
#['CallerIDNum','41215500331']
        }
        case "Newexten" {
#['Event','Newexten']
#['Privilege','dialplan,all']
#['Channel','SIP/41215500331-00000011']
#['Context','fromaccount']
#['Extension','0215502419']
#['Uniqueid','91.121.122.64-1324369709.17']
#['Application','Set']
#['AppData','CDR(CustomerType)=SipAuth']
#['Priority','2']
          if (defined $filtered_event_hash{'Application'} && $filtered_event_hash{'Application'} eq 'Set' && defined $filtered_event_hash{'AppData'}) {
            my $channel = $filtered_event_hash{'Channel'};
            last if (!defined $channels{$channel}{UNIQUEID});
            $channels{$channel}{'DNID'} = $1 if (defined $filtered_event_hash{'Extension'} && $filtered_event_hash{'Extension'} =~ /^(\d+)$/);
            my ( $_var, $_val ) = split(/=/, $filtered_event_hash{'AppData'});
            print 'Newexten Set $channels{'.$filtered_event_hash{'Channel'}.'}{'.$_var.'} = '.$_val."\n" if (DEBUG > 3);
            if($_var eq 'CDR(accountcode)') {
              $channels{$channel}{'ACCOUNTCODE'} = $_val if ( $_val ne '');
            } else {
              $channels{$channel}{$_var} = $_val;
              if ( $_var eq 'H323_ID' && $_val ne '') {
                syslog('debug', "H323 dec: ".$channels{$channel}{'H323_ID'}) if (DEBUG > 2);
                $channels{$channel}{'H323_ID'}=sprintf("%08X %08X %08X %08X",$1,$2,$3,$4) if $channels{$channel}{'H323_ID'} =~ /(\d+)-(\d+)-(\d+)-(\d+)/;
                if (DEBUG) {
                  syslog('debug', "H323 hex: ".$channels{$channel}{'H323_ID'});
                  print $channels{$channel}{'H323_ID'}." - $channel\n" if (DEBUG);
                }
              }
            }
          }
        }
        case "NewAccountCode" {
#['Event','NewAccountCode']
#['Privilege','call,all']
#['Channel','SIP/91.121.151.75-00000014']
#['Uniqueid','91.121.122.64-1324372618.20']
#['Channel2','SIP/91.121.151.75-00000014']
#['Bridgestate','Link']
        }
        case "Link" {
#Link does not seem to appear in the new version of asterisk
#while ( my ($key, $value) = each(%filtered_event_hash) ) {
#   print "$key => $value\n";
#}
        }
        case "VarSet" {
#Event => VarSet
#Uniqueid => 91.121.122.64-1343979015.97
#Privilege => dialplan,all
#Channel => SIP/41225504011-00000061
#Variable => SIPURI
#Value => 1059956257@192_168_1_51
##ATTENTION!!: Asterisk will show the value in the wrong variable in AMI when the variable is set using AGI. Every variable set here appears wrong. This doe snot include variables set by asterisk like QoS values.
          if (defined $filtered_event_hash{'Value'} && defined $filtered_event_hash{'Variable'} && defined $channels{$filtered_event_hash{'Channel'}}) {
            last if (!defined $channels{$filtered_event_hash{'Channel'}}{UNIQUEID});
            $channels{$filtered_event_hash{'Channel'}}{$filtered_event_hash{'Variable'}} = $filtered_event_hash{'Value'};
            print 'VarSet $channels{'.$filtered_event_hash{'Channel'}.'}{'.$filtered_event_hash{'Variable'}.'} = '.$filtered_event_hash{'Value'}."\n" if (DEBUG > 3);
          }
        }
        case "Bridge" {
#Event => Bridge
#Privilege => call,all
#Channel1 => SIP/41225504011-0000005f
#Uniqueid2 => 91.121.122.64-1343978405.96
#CallerID1 => 0041225504011
#Bridgetype => core
#CallerID2 => *0215661380
#Uniqueid1 => 91.121.122.64-1343978404.95
#Value => SIP/212.249.15.5-00000060

          if (defined $filtered_event_hash{'Value'}) {
            my $channel1 = $filtered_event_hash{'Channel1'};
            my $channel2 = $filtered_event_hash{'Value'};
            if (!defined $channels{$channel1}{'H323_ID'} && defined $channels{$channel2}{'H323_ID'}) {
              $channel2 = $filtered_event_hash{'Channel1'};
              $channel1 = $filtered_event_hash{'Value'};
            }
            #Condition bellow prevents sending a new Start Answer and modification of answer time in case of reinvite.
            if ($channels{$channel1}{'STARTSENT'} == 0) {
              $channels{$channel1}{'ANSWERED'}    = 1;
              $channels{$channel1}{'LINK_START'}  = time();
              $channels{$channel2}{'LINK_START'}  = time();
              $channels{$channel1}{'CHANNEL2'}    = $channel2;
              $channels{$channel2}{'CHANNEL2'}    = $channel1;
              $channels{$channel1}{'CALL_ORIGIN'} = 'answer';
              $channels{$channel1}{'STARTSENT'}   = 1;
              print 'Call '.$channel1." answered. Sending ".$channels{$channel1}{'CALL_ORIGIN'}."-START Radius packet.\n" if (DEBUG);
              send_acc('Start','0',%{$channels{$channel1}});
#              send_acc('Start','1',%{$channels{$channel1}}) if ($channels{$channel1}{'XFOLLOW_ME'} eq '1');
            }
          }
        }
        case "Unlink" {
#Event => Unlink
#Privilege => call,all
#Channel1 => SIP/41225504011-00000063
#Uniqueid1 => 91.121.122.64-1343979588.99
#Uniqueid2 => 91.121.122.64-1343979588.100
#CallerID1 => 0041225504011
#CallerID2 => *0215661380
#Value => 6
          my $channel = $filtered_event_hash{'Channel1'};
          last unless $channels{$channel};
          $channels{$channel}{'LINK_END'} = time();
          $channels{$channels{$channel}{'CHANNEL2'}}{'LINK_END'} = time() if defined $channels{$channel}{'CHANNEL2'};
        }
        case "Hangup" {
#Event => Hangup
#Privilege => call,all
#Channel => SIP/212.249.15.5-00000054
#Cause => 16
#ConnectedLineName => 0041225504011
#ConnectedLineNum => 0041225504011
#CallerIDName => <unknown>
#Uniqueid => 91.121.122.64-1343977618.84
#CallerIDNum => *0215661380
#Cause-txt => Normal Clearing
#Value => ANSWER

#Event => Hangup
#Privilege => call,all
#Channel => SIP/41225504011-00000053
#Cause => 16
#ConnectedLineName => <unknown>
#ConnectedLineNum => <unknown>
#CallerIDName => 0041225504011
#Uniqueid => 91.121.122.64-1343977617.83
#CallerIDNum => 0041225504011
#Cause-txt => Normal Clearing
          my $channel2 = $filtered_event_hash{'Channel'};
          my $channel  = $channel2;
          $channel     = $channels{$channel}{'CHANNEL2'} if (!defined $channels{$channel}{'CDR(CustomerType)'} && defined $channels{$channel}{'CHANNEL2'});
          last if (!defined $channels{$channel});
          $filtered_event_hash{'Cause'} = 16 if ($filtered_event_hash{'Cause'} eq '');
          if (!defined $channels{$channel}{'H323_ID'} || !defined $channels{$channel}{'COUNT'}) {
            print "Deleted ".$channel."\n" if (DEBUG);
            delete $channels{$channel};
            last;
          }
          $channels{$channel}{'CALL_END'} = time();
          $channels{$channel}{'CAUSE'} = 0;
          $channels{$channel}{'CAUSE'} = $filtered_event_hash{'Cause'};
          $channels{$channel}{'COUNT'}-- if ($channels{$channel}{'COUNT'} > 0);
          if ($channel ne $channel2 || $channels{$channel}{'DIALED'} == 0) {
            $channels{$channel}{'CALL_ORIGIN'} = 'answer';
            print "Connected channels ".$channels{$channel}{'COUNT'}."\n" if (DEBUG > 1);
            print 'Call '.$channel.">".$channel2." disconnected. Sending ".$channels{$channel}{'CALL_ORIGIN'}."-STOP Radius packet.\n" if (DEBUG);
            send_acc('Stop','0',%{$channels{$channel}});
            $channels{$channel}{'CALL_ORIGIN'} = 'originate';
            print 'Call '.$channel.">".$channel2." disconnected. Sending ".$channels{$channel}{'CALL_ORIGIN'}."-STOP Radius packet.\n" if (DEBUG);
            send_acc('Stop','0',%{$channels{$channel}});
            send_acc('Stop','1',%{$channels{$channel}}) if ($channels{$channel}{'XFOLLOW_ME'} eq '1');
          }
          if (defined $channels{$channel2} && $channel ne $channel2) {
            delete $channels{$channel2}{'CHANNEL1'};
            delete $channels{$channel2};
            print "Deleted ".$channel2."\n" if (DEBUG);
          }
          if ($channels{$channel}{'COUNT'} < 1) {
            print "Deleted ".$channel." (".$channels{$channel}{'COUNT'}.")\n" if (DEBUG);
            delete $channels{$channel}{'CHANNEL2'};
            delete $channels{$channel};
          }
        }
        case "Dial" {
#Event => Dial
#Privilege => call,all
#DestUniqueID => 91.121.122.64-1343977888.90
#SubEvent => Begin
#Channel => SIP/41225504011-00000059
#Destination => SIP/212.249.15.5-0000005a
#Dialstring => 0041215661380@212.249.15.5
#ConnectedLineName => <unknown>
#ConnectedLineNum => <unknown>
#CallerIDName => 0041225504011
#CallerIDNum => 0041225504011
#UniqueID => 91.121.122.64-1343977887.89

#Event => Dial
#Privilege => call,all
#SubEvent => End
#Channel => SIP/41225504011-00000059
#UniqueID => 91.121.122.64-1343977887.89
          if ($filtered_event_hash{'SubEvent'} eq 'End') {
            print "End dial\n" if (DEBUG > 1);
            last;
          }
          my $channel1 = $filtered_event_hash{'Channel'};
          my $channel2 = $filtered_event_hash{'Destination'};
          syslog('debug', "DIAL: $channel1 -> $channel2") if (DEBUG > 1);
          print "DIAL: $channel1 -> $channel2 ".$filtered_event_hash{'Dialstring'}."\n" if (DEBUG);
          if (!defined $channels{$channel1}{'H323_ID'} && defined $channels{$channel2}{'H323_ID'}) {
            $channel1 = $filtered_event_hash{'Destination'};
            $channel2 = $filtered_event_hash{'Channel'};
          }
          $channels{$channel1}{'VLEG'} = $channel2;
          $channels{$channel1}{'DIALED'} = 1;
          $channels{$channel1}{'CALL_ORIGIN'} = 'originate';
#          $channels{$channel2}{'ACCOUNTCODE'} = $channels{$channel1}{'ACCOUNTCODE'};
          $channels{$channel1}{'CALL_ID'}     = $filtered_event_hash{'UniqueID'} if !defined $channels{$channel1}{'CALL_ID'};
#          $channels{$channel2}{'CALL_ID'}     = $channels{$channel1}{'CALL_ID'};
#          $channels{$channel2}{'CALLERID'}    = $channels{$channel1}{'CALLERID'};
#          $channels{$channel2}{'DNID'}        = $channels{$channel1}{'VDNID'};
#          $channels{$channel2}{'H323_ID'}     = $channels{$channel1}{'H323_ID'};
          $channels{$channel1}{'LINK_START2'} = time();
#          $channels{$channel2}{'LINK_START2'} = time();
          $channels{$channel1}{'CHANNEL2'}    = $channel2;
          $channels{$channel2}{'CHANNEL2'}    = $channel1;
          $channels{$channel1}{'COUNT'}++;
          print "Dialed ".$channel1.">".$channel2." (".$channels{$channel1}{'COUNT'}.")\n";
          #print "Sending ".$channels{$channel1}{'CALL_ORIGIN'}."-START Radius packet.\n" if (DEBUG);
          #There should be a stop originate for this start even if the call isnt answered or else the call will have a huge length in CDR
          send_acc('Start','0',%{$channels{$channel1}});
        }
        case "Shutdown" {
          syslog('info', "Asterisk Shutdown");
          safe_exit();
        }
        case "Reload" {
          syslog('info', "Asterisk Reload");
          load_config();
        }
      }
    }
    undef @filtered_event;
    undef %filtered_event_hash;
    shift(@event_pos);
  }
  return;
}
##########################################################################
# 
sub get_billing_addr {
    my $name = shift;
    my @i_billings;
    my @ip_bills;
    my $dbh = DBI->connect("DBI:mysql:$DB_NAME_Rad;host=$DB_HOST_Rad", $DB_USER_Rad, $DB_PASS_Rad);
    my $request = "SELECT i_bill_acc, 1 AS found FROM sipdevices WHERE name='$name' UNION SELECT '1,2' AS i_bill_acc,0 AS found ORDER BY found desc LIMIT 1;";
    print "$request\n" if (DEBUG > 4);
    my $sth = $dbh->prepare($request);
    $sth->execute();
    my $id_bill = $sth->fetchrow_hashref();
    my $i_bill_accs = '1,2';
    $i_bill_accs = $id_bill->{i_bill_acc} if (defined $id_bill->{i_bill_acc});
    print "Result was found? ".$id_bill->{found}."' Sending Radius message to '".$id_bill->{i_bill_acc}."'\n"  if (DEBUG > 1);
    @i_billings = split(',',$i_bill_accs);
    for my $bill (0 .. $#i_billings){
        if ( $i_billings[$bill] == '1' ){
            $ip_bills[$bill] = $rad_serv;
        } elsif ( $i_billings[$bill] == '2' ) {
            $ip_bills[$bill] = $rad_serv2;
        }
    }
    @ip_bills = ($rad_serv,$rad_serv2) if (!@ip_bills);
    return @ip_bills;
}

##########################################################################
# 
sub send_acc {
    my ($acc_type,$followme,%cdr) = @_;
    my $remote;
    my $cause       = "null";
    my $duration    = 0;
    my $date        = 0;
    my $accountcode = $cdr{'ACCOUNTCODE'};
    my $h323confid  = $cdr{'H323_ID'};
    my $CSI         = $cdr{'DNID'};

    if (DEBUG > 2) { 
      while ( my ($key, $value) = each(%channels) ) {
        while ( my ($key2, $value2) = each(%$value) ) {
          if (DEBUG > 3) {
           print "   $key => $key2 - $value2\n";
          } else {
           print "   $key => $key2 - $value2\n" if ($key2 eq 'COUNT');
          }
        }
      }
    }

    if (($acc_type ne 'Start' && $acc_type ne 'Stop') || ($cdr{'H323_ID'} =~ /(\d+)-(\d+)-(\d+)-(\d+)/) || !defined $cdr{'CDR(CustomerType)'} || !defined $accountcode) {
      return;
    }
    $cdr{'CALL_ORIGIN'} = 'originate' if( !defined $cdr{'CALL_ORIGIN'} );
    $cdr{'LINK_START2'} = $cdr{'LINK_END'} if (!defined $cdr{'LINK_START2'});

    if ($followme eq '1') {
      $accountcode = $cdr{'XBILL_TO'} if (defined $cdr{'XBILL_TO'} && $followme eq '1');
      $h323confid  = reverse $cdr{'H323_ID'};
      $CSI = $cdr{'VDNID'};
    } else {
      $CSI = $cdr{'PBCLD'} if ($cdr{'CALL_ORIGIN'} eq "originate" && $cdr{'PBCLD'} ne '');
#      $CSI = $cdr{'VDNID'} if ($cdr{'CALL_ORIGIN'} eq "originate");
    }
    syslog('info', $h323confid." - ".$acc_type."-".$cdr{'CALL_ORIGIN'}." ".$accountcode." > ".$CSI." FM:".$cdr{'XFOLLOW_ME'} ) if (DEBUG);

    # Choose the account that we will check on wich billing we have to send the data
    my $account_res_bill;
    if ( $cdr{'CONTEXT'} eq 'fromaccount' || (defined $cdr{'XBILL_TO'} && $followme eq '1')){
        $account_res_bill = $accountcode;
    } else {
        $account_res_bill = $CSI;
    }
    my @billings=get_billing_addr($account_res_bill);
    foreach my $radius_ip (@billings){
        my $r = new Authen::Radius(Host => $radius_ip, Secret => $cdr{'RADIUS_Secret'}, TimeOut => 10, Service => 'radius-acct');
        if( !defined $r ) {
            syslog('crit', "RADIUS host $radius_ip ERROR");
            print "RADIUS host ".$radius_ip." ERROR\n" if (DEBUG);
            return;
        }

        $r->clear_attributes();
        print $h323confid." - Sending Radius ".$cdr{'CALL_ORIGIN'}." ".$acc_type." ".$cdr{'CALLERID'}.">".$CSI." to Billing $radius_ip\n" if (DEBUG);
        $r->add_attributes (
              { Name => 'Acct-Status-Type',   Value => $acc_type },
              { Name => 'h323-call-origin',   Value => $cdr{'CALL_ORIGIN'} },
              { Name => 'h323-call-type',     Value => $cdr{'CALL_TYPE'} },
              { Name => 'Cisco-AVPair',       Value => "h323-session-protocol=".$cdr{'CALL_PROTOCOL'} },
              { Name => 'NAS-IP-Address',     Value => $cdr{'NAS_IP_Address'} },
              { Name => 'NAS-Port',           Value => $cdr{'NAS-Port'} },
              { Name => 'User-Name',          Value => $accountcode },
              { Name => 'Calling-Station-Id', Value => $cdr{'CALLERID'} },
              { Name => 'Called-Station-Id',  Value => $CSI },
              { Name => 'Cisco-AVPair',       Value => "call-id=".$cdr{'CALL_ID'} },
              { Name => 'Cisco-AVPair',       Value => "Acct-Session-Id=".$cdr{'CALL_ID'} },
              { Name => 'Cisco-AVPair',       Value => "Acct-Delay-Time=0" },
              { Name => 'h323-setup-time',    Value => format_date_gmt($cdr{'CALL_START'}) },
              { Name => 'Cisco-AVPair',       Value => "h323-conf-id=".$h323confid }
        );
        if ($acc_type eq 'Stop') {
            #   TODO why?
            $cause = $cdr{'CAUSE'}-6;
            $cause = 0 if $cause<0;
            $r->add_attributes (
              { Name => 'h323-disconnect-time', Value => format_date_gmt($cdr{'LINK_END'}) },
              { Name => 'Acct-Session-Time', Value => $cdr{'LINK_END'} - $cdr{'LINK_START'} },
              { Name => 'h323-disconnect-cause', Value => $cause }
            #   TODO h323-voice-quality
            );
            $r->add_attributes ( { Name => 'h323-ivr-out', Value => 'PortaBilling_Session:unlock' } ) if ($cdr{'CALL_ORIGIN'} eq "answer");
            $duration=$cdr{'LINK_END'} - $cdr{'LINK_START'};
            if ($followme) {
              $r->add_attributes ( { Name => 'h323-incoming-conf-id', Value => $cdr{'H323_ID'} } );
              #New-Billing format
              $r->add_attributes ( { Name => 'h323-ivr-out', Value => 'PortaBilling_RatePattern:'.$cdr{'XRATE_PTRN'} } ) if (defined $cdr{'XRATE_PTRN'} && $cdr{'XRATE_PTRN'} ne '');
              $r->add_attributes ( { Name => 'h323-ivr-out', Value => 'PortaBilling_AccessCode:FOLLOWME' } );
              #Old-Billing format
              $r->add_attributes ( { Name => 'h323-ivr-out', Value => 'PortaBilling_Auth_CLD:'.$cdr{'XRATE_PTRN'} } ) if (defined $cdr{'XRATE_PTRN'} && $cdr{'XRATE_PTRN'} ne '');
              $r->add_attributes ( { Name => 'h323-ivr-out', Value => 'PortaBilling_Original_CLD:FOLLOWME' } );
            } else {
              $r->add_attributes ( { Name => 'h323-ivr-out', Value => 'PortaBilling_Auth_CLD:'. $cdr{'PBCLD'}} ) if (defined $cdr{'PBCLD'} && $cdr{'PBCLD'} ne '');
            }
        }
        $remote = $cdr{'NAS_IP_Address'};
        if ($cdr{'CALL_ORIGIN'} eq "originate") {
          if (!defined $cdr{'ROUTEIP'} || $cdr{'ROUTEIP'} eq $cdr{'NAS_IP_Address'}) {
            $remote = 'sip-ua';
          } else {
            $remote = $cdr{'ROUTEIP'};
          }
          #The maximum duration of the call is on the wrong field to an Asterisk bug which writes the value of a variable on the previous event when the variable is set from an AGI
          if (defined $cdr{'Dial_Info'}) {
            $cdr{'Dial_Info'} =~ m/L\((\d+):/;
            $r->add_attributes ( { Name => 'h323-ivr-out',      Value => 'DURATION:'. substr($1, 0, -3) } ) if defined $1; 
          }
          $r->add_attributes ( { Name => 'h323-connect-time', Value => format_date_gmt($cdr{'LINK_START'}) } );
          #TODO alert-timepoint when call starts ringing if start
        } else {
          $remote = $cdr{'RECVIP'};
          $r->add_attributes ( { Name => 'h323-connect-time', Value => format_date_gmt($cdr{'LINK_START'}) } );
          $r->add_attributes ( { Name => 'Cisco-AVPair',   Value => "alert-timepoint=".format_date_gmt($cdr{'LINK_START'}) } );
        }
        $r->add_attributes ( { Name => 'h323-remote-address', Value => $remote } );
        $date=time;
            
        if ($acc_type eq 'Start') {
          $r->send_packet (ACCOUNTING_REQUEST);
        } else {
          $r->send_packet (ACCOUNTING_REQUEST) and my $type = $r->recv_packet;
          if (!defined $type) {
            my %data;
            #TODO add the variables needed in case of follow me to second message
            $data{'H323_setup_time'}       = format_date_sql($cdr{'CALL_START'});
            $data{'H323_connect_time'}     = format_date_sql($cdr{'LINK_START'});
            $data{'H323_disconnect_time'}  = format_date_sql($cdr{'LINK_END'});
            $data{'Acct_Session_Time'}     = $cdr{'LINK_END'} - $cdr{'LINK_START'};
            $data{'H323_disconnect_cause'} = $cause;
            $data{'nas_ip_address'}        = $cdr{'NAS_IP_Address'};
            $data{'user_name'}             = $accountcode;
            $data{'call_id'}               = $cdr{'CALL_ID'};
            $data{'H323_conf_id'}          = $h323confid;
            $data{'H323_remote_address'}   = $remote;
            $data{'call_profile'}          = "";
            $data{'call_profile'}          = $cdr{'CDR(CustomerType)'} if defined($data{'CDR(CustomerType)'});
            $data{'creation_date'}         = format_date_sql($date);
            $data{'reception_date'}        = format_date_sql(0);
            $data{'last_send'}             = format_date_sql(time+1);
            $data{'send_counter'}          = 1;
            $data{'status'}                = 0;
            $data{'radius_ip'}             = $radius_ip;
            $data{'Calling-Station-Id'}    = $cdr{'CALLERID'};
            $data{'Called-Station-Id'}     = $CSI;
            $data{'reception_date'}        = 0;
            $data{'STATUS'}                = 0;
            syslog('crit', "Ast-Rad: No response from RADIUS server !");
            print "Ast-Rad: No response from RADIUS server!\n" if DEBUG;
            #TODO sometimes it adds several times the same call
            Failed_radius_packet(%data);
          } else {
            print $h323confid." - Good response from Billing $radius_ip\n" if DEBUG;
            syslog('crit', "Ast-Rad: Good response from RADIUS server");
          }
        } 
   }
}

##########################################################################
#                                 main()                                 #
##########################################################################
# Load Radius dictionary
Authen::Radius->load_dictionary;
if (!load_config()) {
  syslog('crit', "Ast-Rad: Config file error!");
  exit;
}
while ( 1 ) {
  if (!AMI_connect()) {
    syslog('crit', "Ast-Rad: Error in Asterisk Manager connection");
    sleep 1; 
  }
}
##########################################################################
#                                  END                                   #
##########################################################################
